package org.processmining.visualpetrinetanalyzer.algorithms;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Stroke;
import java.util.AbstractMap.SimpleEntry;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.processmining.visualpetrinetanalyzer.models.Description;
import org.processmining.visualpetrinetanalyzer.models.Event;
import org.processmining.visualpetrinetanalyzer.models.VisualPetrinetModel;

public class TreeMap {

	public static List<SimpleEntry<Shape, Description>> Draw(Graphics2D ourGraphics, VisualPetrinetModel model, int panelWidth, int panelHeight, int fontSize) {

		List<SimpleEntry<Shape, Description>> rectangles = new ArrayList<SimpleEntry<Shape, Description>>();
        List<SimpleEntry<Description, Double>> sortedTimes = new ArrayList<SimpleEntry<Description, Double>>();

		for(Event node : model.getNodes()) {
			double eventTime = model.GetValue(node);
			
			if(eventTime > 0) {
				switch(model.axis) {
					case Normal:
						sortedTimes.add(new SimpleEntry<Description, Double>(model.getDescription(node), eventTime));
						break;
					case Log:
						sortedTimes.add(new SimpleEntry<Description, Double>(model.getDescription(node), Math.log(eventTime)));
						break;
				}
			}
		}
		
		Collections.sort(sortedTimes, new Comparator<SimpleEntry<Description, Double>>() {
			public int compare(SimpleEntry<Description, Double> o1, SimpleEntry<Description, Double> o2) {
				return o2.getValue().compareTo(o1.getValue());
			}
		});

		double totalValue = 0;
		for (SimpleEntry<Description, Double> value : sortedTimes) {
			totalValue += value.getValue();
		}
		
		List<SimpleEntry<Description, Rectangle>> treeMap = calculateTreeMap(sortedTimes, totalValue, panelWidth, panelHeight);
		
		if(treeMap.size() > 0) {
	 		Rectangle biggest = treeMap.get(0).getValue();
			double largestSize = biggest.getHeight() * biggest.getWidth();
			
			for(SimpleEntry<Description, Rectangle> keypair : treeMap) {
				Description description = keypair.getKey();
				Rectangle rectangle = keypair.getValue();
				
				double percentage = (rectangle.getHeight() * rectangle.getWidth()) / largestSize * 100;
				percentage = Math.min(100, percentage);
				percentage = Math.max(0, percentage);
				
				double red = (255 * percentage) / 100;
				double green = (255 * (100 - percentage)) / 100;
				
				ourGraphics.setColor(new Color((int)red, (int)green, 0));				
				//ourGraphics.fillRect(rectangle.x, rectangle.y, rectangle.width, rectangle.height);	
				Rectangle rectangleShape = new Rectangle(rectangle.x, rectangle.y, rectangle.width, rectangle.height);	
				ourGraphics.fill(rectangleShape);
				rectangles.add(new SimpleEntry<Shape, Description>(rectangleShape, description));			
				
				ourGraphics.setColor(Color.BLACK);
				Stroke oldStroke = ourGraphics.getStroke();
				ourGraphics.setStroke(new BasicStroke(2));
				ourGraphics.drawRect(rectangle.x, rectangle.y, rectangle.width, rectangle.height);
				
				ourGraphics.setStroke(oldStroke);					

				int bestFontSize = fontSize;
				int biggestWidth = 0;
				boolean bestFontFits = false;
				for (String line : description.getAllTextItems()) {
					int rectFont = fontSize;
					boolean fontFits = false;
					while(!fontFits && rectFont > (1 / 1000000)) {
						Font font = new Font("TimesRoman", Font.PLAIN, rectFont);
						FontMetrics metrics = ourGraphics.getFontMetrics(font);
						int hgt = metrics.getHeight();
						int adv = metrics.stringWidth(line);
						
						if(adv < rectangle.width) {
							fontFits = true;
							bestFontFits = true;
							if(rectFont < bestFontSize) {
								bestFontSize = rectFont;
							}
							if(adv > biggestWidth) { 
								biggestWidth = adv;
							}
						}
						rectFont *= 0.8;
					}
				}

				if(bestFontFits) {
					Font font = new Font("TimesRoman", Font.BOLD, bestFontSize);
					FontMetrics metrics = ourGraphics.getFontMetrics(font);
					ourGraphics.setColor(Color.BLACK);
					ourGraphics.setFont(font);
					
					String header = description.getHeader();
					int hgt = metrics.getHeight();
					int adv = metrics.stringWidth(header);

					int totalHeight = hgt * 8;
					if (totalHeight <= rectangle.height) {
						int lineOffset = -(totalHeight / 2);
						ourGraphics.drawString(header, rectangle.x + (rectangle.width / 2) - (biggestWidth / 2), rectangle.y + (rectangle.height / 2) + lineOffset);
						lineOffset += hgt;
	
						font = new Font("TimesRoman", Font.PLAIN, bestFontSize);
						metrics = ourGraphics.getFontMetrics(font);
						ourGraphics.setFont(font);
						for (String line : description.getItems()) {
							hgt = metrics.getHeight();
							adv = metrics.stringWidth(line);
							
							ourGraphics.drawString(line, rectangle.x + (rectangle.width / 2) - (biggestWidth / 2), rectangle.y + (rectangle.height / 2) + lineOffset);
							lineOffset += hgt;
						}
					}
				}
			}
		}
		
		return rectangles;
	}
	
	private static List<SimpleEntry<Description, Rectangle>> calculateTreeMap(List<SimpleEntry<Description, Double>> sortedTimes, double totalValue, int panelWidth, int panelHeight) {
		double remainingWidth = panelWidth;
		double remainingHeight = panelHeight;
		
		List<SimpleEntry<Description, Rectangle>> results = new ArrayList<SimpleEntry<Description, Rectangle>>();
		int currentIndex = 0;
		double prevRatio = Double.MAX_VALUE;
		
		double cacheSize = 0;
		List<SimpleEntry<Description, Double>> cacheList = new ArrayList<SimpleEntry<Description, Double>>();
		
		while(currentIndex < sortedTimes.size()) {

			SimpleEntry<Description, Double> value = sortedTimes.get(currentIndex);
			double relativeSize = (double)panelWidth * (double)panelHeight * (value.getValue() / totalValue);
			
			if(drawVertically(remainingWidth, remainingHeight)) {
			
				double width = (relativeSize + cacheSize) / remainingHeight;
				double ratio = getAspectRatio(width, relativeSize / width);
				
				if(ratio < prevRatio) {
					
					cacheList.add(new SimpleEntry<Description, Double>(value.getKey(), relativeSize));
					cacheSize += relativeSize;
					prevRatio = ratio;
					currentIndex++;
				} else {						
					double cacheWidth = cacheSize / remainingHeight;
					double handledHeight = 0;
					for(SimpleEntry<Description, Double> cacheValue : cacheList) {
						double height = cacheValue.getValue() / cacheWidth;
						results.add(new SimpleEntry<Description, Rectangle>(cacheValue.getKey(), new Rectangle((int)Math.ceil(panelWidth - remainingWidth), (int)Math.ceil(panelHeight - remainingHeight + handledHeight), (int)Math.ceil(cacheWidth), (int)Math.ceil(height))));
						handledHeight += height;
					}
					
					remainingWidth = remainingWidth - cacheWidth;
					cacheSize = 0;
					cacheList.clear();
					prevRatio = Double.MAX_VALUE;
				}			
			} else {
				
				double height = (relativeSize + cacheSize) / remainingWidth;
				double ratio = getAspectRatio(height, relativeSize / height);
				
				if(ratio < prevRatio) {
					cacheList.add(new SimpleEntry<Description, Double>(value.getKey(), relativeSize));
					cacheSize += relativeSize;
					prevRatio = ratio;
					currentIndex++;
				} else {
					
					double cacheHeight = cacheSize / remainingWidth;
					double handledWidth = 0;
					for(SimpleEntry<Description, Double> cacheValue : cacheList) {
						double width = cacheValue.getValue() / cacheHeight;
						results.add(new SimpleEntry<Description, Rectangle>(cacheValue.getKey(), new Rectangle((int)Math.ceil(panelWidth - remainingWidth + handledWidth), (int)Math.ceil(panelHeight - remainingHeight), (int)Math.ceil(width), (int)Math.ceil(cacheHeight))));
						handledWidth += width;
					}
					
					remainingHeight = remainingHeight - cacheHeight;
					cacheSize = 0;
					cacheList.clear();
					prevRatio = Double.MAX_VALUE;
				}					
			}
		}
		
		//Fill up the last rectangles
		if(drawVertically(remainingWidth, remainingHeight)) {							
			double cacheWidth = cacheSize / remainingHeight;
			double handledHeight = 0;
			for(SimpleEntry<Description, Double> cacheValue : cacheList) {
				double height = cacheValue.getValue() / cacheWidth;
				results.add(new SimpleEntry<Description, Rectangle>(cacheValue.getKey(), new Rectangle((int)Math.ceil(panelWidth - remainingWidth), (int)Math.ceil(panelHeight - remainingHeight + handledHeight), (int)Math.ceil(cacheWidth), (int)Math.ceil(height))));
				handledHeight += height;
			}								
		} else {				
			double cacheHeight = cacheSize / remainingWidth;
			double handledWidth = 0;
			for(SimpleEntry<Description, Double> cacheValue : cacheList) {
				double width = cacheValue.getValue() / cacheHeight;
				results.add(new SimpleEntry<Description, Rectangle>(cacheValue.getKey(), new Rectangle((int)Math.ceil(panelWidth - remainingWidth + handledWidth), (int)Math.ceil(panelHeight - remainingHeight), (int)Math.ceil(width), (int)Math.ceil(cacheHeight))));
				handledWidth += width;
			}	
		}
		
		return results;
	}
	
	private static double getAspectRatio(double x, double y) { 
		return Math.max(x/ y, y / x);
	}
	
	private static boolean drawVertically (double width, double height) { 
	    return width > height; 
	}
}
