package org.processmining.visualpetrinetanalyzer.algorithms;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Shape;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.util.AbstractMap.SimpleEntry;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.ToolTipManager;

import org.processmining.visualpetrinetanalyzer.models.Description;
import org.processmining.visualpetrinetanalyzer.models.VisualPetrinetModel;

public class VisualPetrinetVisualizerAlgorithm {

	PanAndZoomCanvas canvas;
	AffineTransform at;   // the current pan and zoom transform
	Point2D XFormedPoint; // storage for a transformed mouse point
	int containerSize = 100;
	int fontSize = 15;
	int panelWidth = 1500;
	int panelHeight = 1000;
	
	protected JComponent CreateVisualization(VisualPetrinetModel model) {
        canvas = new PanAndZoomCanvas(model);
		PanningHandler panner = new PanningHandler();
		canvas.addMouseListener(panner);
		canvas.addMouseMotionListener(panner);
		canvas.setBorder(BorderFactory.createLineBorder(Color.black));
		canvas.addMouseWheelListener(new MouseWheelListener() {
			public void mouseWheelMoved(MouseWheelEvent e) {
				int rotation = e.getWheelRotation();	
			    canvas.scale -= ((double)(rotation) / 10);
			    if(canvas.scale < 0.25)
			    	canvas.scale = 0.25;
			    if(canvas.scale > 10)
			    	canvas.scale = 10;
			    canvas.repaint();
			}			
		});
		return canvas;
	}
	
	class PanAndZoomCanvas extends JComponent {
		VisualPetrinetModel model;
		double translateX;
		double translateY;
		double scale;
		List<SimpleEntry<Shape, Description>> shapes;
		
		PanAndZoomCanvas(VisualPetrinetModel model) {
			this.model = model;
			this.translateX = 0;
		    this.translateY = 0;
		    this.scale = 1;
		    shapes = new ArrayList<SimpleEntry<Shape, Description>>();
            ToolTipManager.sharedInstance().registerComponent(this);
		}
		
		public void paintComponent(Graphics g) {
		    Graphics2D ourGraphics = (Graphics2D) g;
		    // save the original transform so that we can restore
		    // it later
		    AffineTransform saveTransform = ourGraphics.getTransform();
		
		    // blank the screen. If we do not call super.paintComponent, then
		    // we need to blank it ourselves
		    ourGraphics.setColor(Color.WHITE);
		    ourGraphics.fillRect(0, 0, getWidth(), getHeight());
				
		    // We need to add new transforms to the existing
		    // transform, rather than creating a new transform from scratch.
		    // If we create a transform from scratch, we will
		    // will start from the upper left of a JFrame, 
		    // rather than from the upper left of our component
		    at = new AffineTransform(saveTransform);
		
		    // The zooming transformation. Notice that it will be performed
		    // after the panning transformation, zooming the panned scene,
		    // rather than the original scene
		    at.translate(getWidth()/2, getHeight()/2);
		    at.scale(scale, scale);
		    at.translate(-getWidth()/2, -getHeight()/2);
		
		    // The panning transformation
		    at.translate(translateX, translateY);
		
		    ourGraphics.setTransform(at);

			Font font = new Font("TimesRoman", Font.PLAIN, fontSize);
			ourGraphics.setColor(Color.BLACK);
			ourGraphics.setFont(font);
			
			switch(model.visualization) {
				case TreeMap:
					shapes = TreeMap.Draw(ourGraphics, model, panelWidth, panelHeight, fontSize);
					break;
				case Matrix:
					shapes = Matrix.Draw(ourGraphics, model, panelWidth, panelHeight, fontSize);
					break;
				case Chart:
					shapes = Chart.Draw(ourGraphics, model, panelWidth, panelHeight, fontSize);
					break;
			}
			
		    ourGraphics.setTransform(saveTransform);
		}
		
		public Dimension getPreferredSize() {
		    return new Dimension(panelWidth, panelHeight);
		}

        @Override
        public String getToolTipText(MouseEvent event) {
            Point p = new Point(event.getX(), event.getY());
            if(shapes != null) {
	            for(SimpleEntry<Shape, Description> shape : shapes) {
	            	Description description = tooltipForCircle(p, shape.getKey(), shape.getValue());
	                if (description != null) {
	                    return description.getHTMLTooltip();
	                }
	            }
            }
            return super.getToolTipText(event);
        }

        protected Description tooltipForCircle(Point p, Shape shape, Description tipText) {
        	
		    AffineTransform at = new AffineTransform();
		    at.translate(getWidth()/2, getHeight()/2);
		    at.scale(scale, scale);
		    at.translate(-getWidth()/2, -getHeight()/2);
		    at.translate(translateX, translateY);
		    Shape transformed = at.createTransformedShape(shape);
		    
            if (transformed.contains(p)) {

                return tipText;
            }
            return null;
        }
	}
	
	class PanningHandler implements MouseListener, MouseMotionListener {
		double referenceX;
		double referenceY;
		AffineTransform initialTransform;
		
		public void mousePressed(MouseEvent e) {
		
		    try {
		    	XFormedPoint = at.inverseTransform(e.getPoint(), null);
		    }
		    catch (NoninvertibleTransformException te) {
		    	System.out.println(te);
		    }
		
		    referenceX = XFormedPoint.getX();
		    referenceY = XFormedPoint.getY();
		    initialTransform = at;
		}
		
		public void mouseDragged(MouseEvent e) {
		    try {
		    	XFormedPoint = initialTransform.inverseTransform(e.getPoint(), null);
		    }
		    catch (NoninvertibleTransformException te) {
		    	System.out.println(te);
		    }
			
		    double deltaX = XFormedPoint.getX() - referenceX;
		    double deltaY = XFormedPoint.getY() - referenceY;
			
		    referenceX = XFormedPoint.getX();
		    referenceY = XFormedPoint.getY();
		    
		    canvas.translateX += deltaX;
		    canvas.translateY += deltaY;
			
		    canvas.repaint();
		}
		
		public void mouseClicked(MouseEvent e) {}
		public void mouseEntered(MouseEvent e) {}
		public void mouseExited(MouseEvent e) {}
		public void mouseMoved(MouseEvent e) {}
		public void mouseReleased(MouseEvent e) {}
	}
}
