package org.processmining.visualpetrinetanalyzer.models;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.visualpetrinetanalyzer.parameters.Axis;
import org.processmining.visualpetrinetanalyzer.parameters.CalculationMethod;
import org.processmining.visualpetrinetanalyzer.parameters.VisualPetrinetModelParameters;
import org.processmining.visualpetrinetanalyzer.parameters.Visualization;

public class VisualPetrinetModel {

	public Visualization visualization;
	public CalculationMethod calculationMethod;
	public Axis axis;
	
	private Map<Event, TimeRecords> TimestampMap;

	public VisualPetrinetModel(Petrinet petrinet, VisualPetrinetModelParameters parameters) {
		visualization = parameters.visualization;
		calculationMethod = parameters.calculationMethod;
		axis = parameters.axis;

	    Collection<Transition> transitions = petrinet.getTransitions();
		
		TimestampMap = new HashMap<Event, TimeRecords>();

		Map<Transition, List<PetrinetNode>> petrinetMap = new HashMap<Transition, List<PetrinetNode>>();
		
		for(Transition transition : transitions) {
			petrinetMap.put(transition, new ArrayList<PetrinetNode>());
		}
		
		Collection<Place> places = petrinet.getPlaces();
		
	    for(PetrinetNode node : places ) {
	    	Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = petrinet.getInEdges(node);
	    	Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = petrinet.getOutEdges(node);

	    	for(PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> outEdge : outEdges) {
		    	PetrinetNode target = outEdge.getTarget();

		    	for(PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inEdge : inEdges) {
			    	PetrinetNode source = inEdge.getSource();
		    		petrinetMap.get(target).add(source);
		    	}
		    }
	    }
	    
	    //Create a copy, must be done like this because the lists must also be a new instance
	    Map<Transition, List<PetrinetNode>> petrinetMapCopy = new HashMap<Transition, List<PetrinetNode>>();
	    for(Transition transition : petrinetMap.keySet()) {
	    	petrinetMapCopy.put(transition, new ArrayList<PetrinetNode>());
			for(PetrinetNode node : petrinetMap.get(transition)) {
				petrinetMapCopy.get(transition).add(node);
			}
	    }
	    
	    for(Transition transition : petrinetMap.keySet()) {
	    	if(transition.isInvisible()) {
	    		
	    		//Remove it because it is a Tau step
	    		for(Transition findTransition : petrinetMap.keySet()) {
	    			// For each transition, check if the source is this tau
	    			for(PetrinetNode source : petrinetMap.get(findTransition)) {
	    				if(source != null && source == transition) {
	    					// If it is this tau, remove it as a source and add the source from the Tau transition
	    					if(petrinetMapCopy.containsKey(findTransition)) {
		    					petrinetMapCopy.get(findTransition).remove(source);
		    					List<PetrinetNode> tauSources = petrinetMapCopy.get(transition);
		    					petrinetMapCopy.get(findTransition).addAll(tauSources);
	    					}
	    				}
	    			}
	    		}
	    		
	    		petrinetMapCopy.remove(transition);
	    	}
	    }
	    petrinetMap = petrinetMapCopy;

	    //Create the timestamp map for all target-source combinations
	    for(Transition targetNode : petrinetMap.keySet()) {
	    	String target = targetNode.getLabel();
	    	Event targetEvent = getEventFromLabel(target);
			TimestampMap.put(targetEvent, new TimeRecords(targetEvent));
    		for(PetrinetNode sourceNode : petrinetMap.get(targetNode)) {
    			TimestampMap.get(targetEvent).AddSource(getEventFromLabel(sourceNode.getLabel()));
    		}
	    }
	}
	
	private Event getEventFromLabel(String label) {
		int p = label.lastIndexOf("+");
		if(p >= 0) {
			String transition = label.substring(p + 1);
			String name = label.substring(0, p);
			return new Event(name, transition);
		} else {
			return new Event(label, "complete");
		}
	}

	public boolean targetHasSource(Event target, Event source) {
		if(TimestampMap.containsKey(target) && TimestampMap.get(target).HasSource(source))
			return true;
		else
			return false;
	}
	
	public void addTimestamp(Event target, Event source, long execTime) {
		if(TimestampMap.containsKey(target)) {
			TimestampMap.get(target).AddSourceTime(source, execTime);
		}
	}
	
	public void IncreaseFrequency(Event target) {
		if(TimestampMap.containsKey(target)) {
			TimestampMap.get(target).IncreaseFrequency();
		}
	}

	public Set<Event> getNodes() {
		return TimestampMap.keySet();
	}
	
	public double GetValue(Event target) {
		if(TimestampMap.containsKey(target)) {
			TimeRecords record = TimestampMap.get(target);
			
			switch(calculationMethod) {
				case Frequence:
					return record.GetFrequency();
				case PredFrequence:
					return record.GetPredFrequency();
				case Sum:
					return record.getSum();
				case Average:
					return record.getAverage();
				case Max:
					return record.getMax();
				case Min:
					return record.getMin();
				case StdDev:
					return record.getStdDev();

			}
		}
		return 0;
	}
	
	public double GetValue(Event target, Event source) {
		if(TimestampMap.containsKey(target)) {
			TimeRecords record = TimestampMap.get(target);
			
			switch(calculationMethod) {
				case Frequence:
					return record.GetFrequency();
				case PredFrequence:
					return record.GetPredFrequency(source);
				case Sum:
					return record.getSum(source);
				case Average:
					return record.getAverage(source);
				case Max:
					return record.getMax(source);
				case Min:
					return record.getMin(source);
				case StdDev:
					return record.getStdDev(source);
			}
		}
		return 0;
	}
	
	public double GetHighestValue() {
		double highest = 0;
		for(Event target : TimestampMap.keySet()) {
			double targetValue = GetValue(target);
			if(targetValue > highest)
				highest = targetValue;
		}
		return highest;
	}
	
	public double GetHighestValuePerSource() {
		double highest = 0;
		for(Event target : TimestampMap.keySet()) {
			for(Event source : TimestampMap.get(target).GetSource()) {
				double targetValue = GetValue(target, source);
				if(targetValue > highest)
					highest = targetValue;
			}
		}
		return highest;		
	}

	public Description getDescription(Event target) {
		if(TimestampMap.containsKey(target)) {
			return TimestampMap.get(target).getDescription();
		}
		return null;
	}

	public Description getDescription(Event target, Event source) {
		if(TimestampMap.containsKey(target)) {
			return TimestampMap.get(target).getDescription(source);
		}
		return null;
	}
}
